from datetime import datetime

import numpy as np
from rich.console import Console
from rich.tree import Tree

console = Console()


def print_creator(color):
    """
    Returns a function that prints text in the specified color.

    :param color: The color to use for printing.
    :type color: str
    :return: A function that prints text in the specified color.
    :rtype: function
    """

    def print_creator_inner(*args, **kwargs):
        console.print(*args, **kwargs, style=color)

    return print_creator_inner


red = print_creator("red")
green = print_creator("green")
bold = print_creator("bold")
blue = print_creator("blue")
purple = print_creator("purple")
orange = print_creator("orange")


def inspect(obj, list_expand=3):
    """
    Inspects the given object and prints a tree representation of its attributes and values.

    :param obj: The object to inspect.
    :type obj: object
    :param list_expand: The maximum number of elements to expand in a list or numpy array, defaults to 3.
    :type list_expand: int, optional
    :return: None
    :rtype: None
    """
    try:
        import torch

        torch_exists = True
    except ImportError:
        torch_exists = False

    console = Console(record=True)

    class TreeWrapper:
        def __init__(self) -> None:
            self.tree = None

        def add(self, label):
            if self.tree is None:
                self.tree = Tree(label)  # type: ignore
                return self.tree
            else:
                return self.tree.add(label)  # type: ignore

    def type_fmt(obj):
        """
        Extracts the type name from an object's type and returns it as a string.

        :param obj: The object whose type name is to be extracted.
        :type obj: Any
        :return: The name of the object's type.
        :rtype: str
        """
        # get string within single quotes regex
        import re

        m = re.search(r"'(.*?)'", str(type(obj)))
        if m is not None:
            return m.group(1)
        return None

    def add_children(obj: object, root: Tree) -> None:
        """
        Recursively adds child nodes to a given root node based on the type of the input object.

        :param obj: The object to add child nodes for.
        :type obj: object
        :param root: The root node to add child nodes to.
        :type root: Tree
        """
        if isinstance(obj, dict):
            dict_node = root.add("dict")
            for key in obj.keys():
                add_children(obj[key], dict_node.add(f"[bold]{key}[/bold]"))
        elif isinstance(obj, list) or isinstance(obj, np.ndarray):
            list_node = root.add(f"{type_fmt(obj)}[{len(obj)}]")
            if len(obj) == 0:
                return
            for i in range(min(list_expand, len(obj))):
                add_children(obj[i], list_node)
            if len(obj) > list_expand:
                list_node.add("...")
        elif torch_exists and isinstance(obj, torch.Tensor):
            root.add(f"tensor[[blue]{tuple(obj.shape)}; {obj.dtype}[/blue]]")
        else:
            repr = str(obj)
            if len(repr) > 100:
                repr = repr[:100] + "..."
            root.add(f"{type_fmt(obj)}: [blue]{repr}[/blue]")

    treewrapper = TreeWrapper()
    add_children(obj, treewrapper)  # type: ignore
    console.print(treewrapper.tree)



def map_recursive(fctn, x_or_iterable):
    """
    Apply `fctn` to each element in x_or_iterable.

    This is a generalization of the map function since this will work
    recursively for iterables.

    :param fctn: Function from element of iterable to something.
    :param x_or_iterable: An element or an Iterable of an element.
    :return: The same (potentially recursive) iterable but with
    all the elements transformed by fctn.
    """
    # if isinstance(x_or_iterable, Iterable):
    if isinstance(x_or_iterable, list) or isinstance(x_or_iterable, tuple):
        return type(x_or_iterable)(map_recursive(fctn, item) for item in x_or_iterable)
    else:
        return fctn(x_or_iterable)


def filter_recursive(x_or_iterable):
    """
    Filter out elements that are Falsy (where bool(x) is False) from
    potentially recursive lists.

    :param x_or_iterable: An element or a list.
    :return: If x_or_iterable is not an Iterable, then return x_or_iterable.
    Otherwise, return a filtered version of x_or_iterable.
    """
    if isinstance(x_or_iterable, list):
        new_items = []
        for sub_elem in x_or_iterable:
            filtered_sub_elem = filter_recursive(sub_elem)
            if filtered_sub_elem is not None and not (
                isinstance(filtered_sub_elem, list) and len(filtered_sub_elem) == 0
            ):
                new_items.append(filtered_sub_elem)
        return new_items
    else:
        return x_or_iterable


def batch(iterable, n=1):
    """
    Split an interable into batches of size `n`. If `n` does not evenly divide
    `iterable`, the last slice will be smaller.

    https://stackoverflow.com/questions/8290397/how-to-split-an-iterable-in-constant-size-chunks

    Usage:
    ```
        for i in batch(range(0,10), 3):
            print i

        [0,1,2]
        [3,4,5]
        [6,7,8]
        [9]
    ```
    """
    l = len(iterable)
    for ndx in range(0, l, n):
        yield iterable[ndx : min(ndx + n, l)]


def find_key_recursive(obj, key):
    """
    Recursively searches for a key in a nested dictionary.

    :param obj: The dictionary to search in.
    :type obj: dict
    :param key: The key to search for.
    :type key: Any
    :return: The value associated with the key if found, otherwise None.
    :rtype: Any
    """
    if key in obj:
        return obj[key]
    for k, v in obj.items():
        if isinstance(v, dict):
            result = find_key_recursive(v, key)
            if result is not None:
                return result



from typing import Any, Optional, Union

import functools

import numpy as np
import torch

from eztils import normalize


def to_np(x):
    """
    Convert a tensor to a numpy array.

    :param x: Input tensor.
    :type x: torch.Tensor
    :return: Numpy array.
    :rtype: numpy.ndarray
    """
    if torch.is_tensor(x):
        x = x.detach().cpu().numpy()
    return x


def to_img(x: torch.Tensor):
    """converts a tensor to a numpy array suitable for saving as an image

    :param x:   a tensor of shape (C, H, W)
    :type x:   torch.Tensor
    :return:   a numpy array of shape (H, W, C) with values in [0, 255]
    :rtype:   np.ndarray
    """
    normalized = normalize(x)
    array = to_np(normalized)
    array = np.transpose(array, (1, 2, 0))
    return (array * 255).astype(np.uint8)


def to_torch(x, dtype=None, device=None):
    """
    Converts a numpy array or a Python scalar to a PyTorch tensor.

    Args:
        x (numpy.ndarray, Python scalar): Input data to be converted to PyTorch tensor.
        dtype (torch.dtype, optional): Data type of the resulting tensor. Defaults to None.
        device (torch.device, optional): Device on which the tensor will be allocated. Defaults to None.

    Returns:
        torch.Tensor: A PyTorch tensor with the same data as the input.

    """
    from eztils.torch import DEVICE, DTYPE

    dtype = dtype or DTYPE
    device = device or DEVICE
    if isinstance(x, dict):
        return {k: to_torch(v, dtype, device) for k, v in x.items()}
    elif torch.is_tensor(x):
        return x.to(device).type(dtype)
    return torch.tensor(x, dtype=dtype, device=device)


def to_device(
    input: Any,
    device: Union[str, torch.device, int] = None,
    inplace: Optional[bool] = True,
) -> Any:
    """
    Recursively places tensors on the appropriate device.

    Args:
        input (Any): The input tensor or collection of tensors to be moved to the device.
        device (Union[str, torch.device, int], optional): The device to move the tensors to. Defaults to None.
        inplace (bool, optional): Whether to perform the operation in place. Defaults to True.

    Returns:
        Any: The tensor or collection of tensors moved to the device.

    Raises:
        NotImplementedError: If the input is of a type that is not supported.
    """
    from eztils.torch import DEVICE

    device = device or DEVICE

    """Recursively places tensors on the appropriate device."""
    if input is None:
        return input
    elif isinstance(input, torch.Tensor):
        return input.to(device)
    elif isinstance(input, tuple):
        return tuple(
            to_device(input=subinput, device=device, inplace=inplace)
            for subinput in input
        )
    elif isinstance(input, list):
        if inplace:
            for i in range(len(input)):
                input[i] = to_device(input=input[i], device=device, inplace=inplace)
            return input
        else:
            return [
                to_device(input=subpart, device=device, inplace=inplace)
                for subpart in input
            ]
    elif isinstance(input, dict):
        if inplace:
            for key in input:
                input[key] = to_device(input=input[key], device=device, inplace=inplace)
            return input
        else:
            return {
                k: to_device(input=input[k], device=device, inplace=inplace)
                for k in input
            }
    elif isinstance(input, set):
        if inplace:
            for element in list(input):
                input.remove(element)
                input.add(to_device(element, device=device, inplace=inplace))
        else:
            return {to_device(k, device=device, inplace=inplace) for k in input}
    elif isinstance(input, np.ndarray) or np.isscalar(input) or isinstance(input, str):
        return input
    elif hasattr(input, "to"):
        # noinspection PyCallingNonCallable
        return input.to(device=device, inplace=inplace)
    else:
        raise NotImplementedError(
            f"Sorry, value of type {type(input)} is not supported."
        )


"""
Recursively places tensors on the CPU.
"""
to_cpu = functools.partial(to_device, device="cpu")



# %%

import os
import signal
import subprocess
import sys
import time
from datetime import datetime
from itertools import product
from pathlib import Path
from subprocess import Popen

from art import tprint
from rich import print
from rich.table import Table


def get_user_defined_attrs(cls) -> list:
    return [
        attr
        for attr in dir(cls)
        if not callable(getattr(cls, attr)) and not attr.startswith("__")
    ]


class BaseHyperParameters:
    @classmethod
    def get_product(cls):
        return list(
            product(*[getattr(cls, attr) for attr in get_user_defined_attrs(cls)])
        )


# %%
def kill_processes(processes):
    for process in processes:
        try:
            # Sends SIGTERM to the entire process group
            os.killpg(os.getpgid(process.pid), signal.SIGTERM)
            # Optionally, you can wait for the processes to ensure they have stopped
            process.wait(timeout=5)
        except Exception as e:
            print(f"Failed to terminate process group {process.pid}: {e}")


def run_parallel(
    hparam_cls: BaseHyperParameters,
    use_cuda_visible_devices: bool = False,
    base_cmd: str = "python3 scripts/eval.py",
    typer_style: bool = True,
    data_path: str = "./runs",
    sleep_time: int = 5,
    debug: bool = False,
) -> list:
    """
    Run parallel processes with different hyperparameters.

    :return: List of spawned processes
    """
    from eztils import datestr

    """
    Handle signals
    """
    processes = []  # Store the process IDs

    # Define your signal handler function
    def signal_handler(signum, frame):
        print(f"Received signal: {signal.Signals(signum).name}")
        kill_processes(processes)
        sys.exit(1)  # or any appropriate exit code

    # Register the handlers within the function
    signal.signal(signal.SIGINT, signal_handler)
    signal.signal(signal.SIGQUIT, signal_handler)
    """
    End handle signals
    """

    hparams = hparam_cls.get_product()
    attrs = get_user_defined_attrs(hparam_cls)
    tprint("Run Parallel", font="bigchief")
    print("Starting at", datestr(), "\n\n")

    table = Table(title="Hyperparameters")

    table.add_column("")
    for attr in attrs:
        table.add_column(attr)

    # Add rows to the table
    for i, values in enumerate(hparams):
        table.add_row(str(i), *[str(value) for value in values])

    # print(pd.DataFrame(hparams, columns=attrs))
    print(table)

    # Assuming d_count and base_cmd are defined elsewhere
    if use_cuda_visible_devices:
        import torch

        d_count = torch.cuda.device_count()

    for i, values in enumerate(hparams):
        args = {attr: value for attr, value in zip(attrs, values)}

        device_id = str(i % d_count) if use_cuda_visible_devices else ""
        cuda_cmd = f"CUDA_VISIBLE_DEVICES={device_id}" if device_id else ""

        # Append hyperparameters to the command
        full_cmd_args = []
        for arg, value in args.items():
            if typer_style:
                hyphen_arg = arg.replace("_", "-")
                if isinstance(value, bool):
                    arg_str = f"--{'no-' if not value else ''}{hyphen_arg}"
                else:
                    arg_str = f"--{hyphen_arg} {value}"
            else:
                arg_str = f"--{arg}={value}"

            full_cmd_args.append(arg_str)

        full_cmd = f"{cuda_cmd} {base_cmd} {' '.join(full_cmd_args)}"
        fout = f'{datetime.now().strftime("%b_%d")}_{i}.out'
        data_path = Path(data_path).resolve()
        data_path.mkdir(exist_ok=True, parents=True)
        output_file = data_path / fout
        print(f"Running {i}: {full_cmd} > {output_file}")

        if not debug:
            time.sleep(sleep_time)
            with open(output_file, "w") as fout:
                process = Popen(
                    full_cmd,
                    stdout=fout,
                    stderr=subprocess.STDOUT,
                    preexec_fn=os.setsid,
                    shell=True,
                )
            processes.append(process)  # this isn't working properly...TODO Fix
            print("pids", " ".join([str(p.pid) for p in processes]))

    return processes


if __name__ == "__main__":

    class HyperParameters(BaseHyperParameters):
        frame_offset = [-1]
        gen_steps = [-1]
        seeds = [12, 13, 42]
        epochs = [300000, 400000, 490000]

    run_parallel(
        HyperParameters,
        use_cuda_visible_devices=False,
        typer_style=True,
        data_path="./data",
        sleep_time=0,
    )



from copy import deepcopy

import torch
from torch import nn


class EMA(nn.Module):
    """Model Exponential Moving Average V2 from timm"""

    def __init__(self, model, decay=0.9999):
        super().__init__()
        # make a copy of the model for accumulating moving average of weights
        self.module = deepcopy(model)
        self.module.eval()
        self.decay = decay

    def _update(self, model, update_fn):
        with torch.no_grad():
            for ema_v, model_v in zip(
                self.module.state_dict().values(), model.state_dict().values()
            ):
                ema_v.copy_(update_fn(ema_v, model_v))

    def update(self, model):
        self._update(
            model, update_fn=lambda e, m: self.decay * e + (1.0 - self.decay) * m
        )

    def set(self, model):
        self._update(model, update_fn=lambda e, m: m)

    def forward(self, *args, **kwargs):
        return self.module.forward(*args, **kwargs)


class LayerNorm(nn.Module):
    """
    Simple 1D LayerNorm.
    """

    def __init__(self, features, center=True, scale=False, eps=1e-6):
        super().__init__()
        self.center = center
        self.scale = scale
        self.eps = eps
        if self.scale:
            self.scale_param = nn.Parameter(torch.ones(features))
        else:
            self.scale_param = None
        if self.center:
            self.center_param = nn.Parameter(torch.zeros(features))
        else:
            self.center_param = None

    def forward(self, x):
        mean = x.mean(-1, keepdim=True)
        std = x.std(-1, keepdim=True)
        output = (x - mean) / (std + self.eps)
        if self.scale:
            output = output * self.scale_param
        if self.center:
            output = output + self.center_param
        return output


class ParallelLayerNorm(LayerNorm):
    """
    A layer normalization module that applies layer normalization across multiple heads.

    Args:
        num_heads (int): The number of heads.
        features (int): The number of features per head.
        center (bool, optional): If True, add learnable bias parameter. Default: True.
        scale (bool, optional): If True, add learnable scale parameter. Default: False.
        eps (float, optional): A value added to the denominator for numerical stability. Default: 1e-6.
    """

    def __init__(self, num_heads, features, center=True, scale=False, eps=0.000001):
        super().__init__(features * num_heads, center, scale, eps)
        self.num_heads = num_heads
        self.features = features

    def forward(self, x_):
        x = x_.reshape(x_.shape[0], self.num_heads, self.features)

        mean = x.mean(-1, keepdim=True)
        std = x.std(-1, keepdim=True)
        output = (x - mean) / (std + self.eps)

        if self.scale:
            output = output * self.scale_param.reshape(self.num_heads, self.features)
        if self.center:
            output = output + self.center_param.reshape(self.num_heads, self.features)

        return output.reshape(*x_.shape)



import torch


def atanh(x):
    """
    Returns the inverse hyperbolic tangent of x.

    :param x: Input tensor.
    :type x: torch.Tensor
    :return: The inverse hyperbolic tangent of x.
    :rtype: torch.Tensor
    """
    # errors or instability at values near 1
    x = torch.clamp(x, -0.999999, 0.999999)
    return 0.5 * (torch.log(1 + x) - torch.log(1 - x))



from typing import List

import errno
import shutil
import subprocess
from collections import namedtuple
from pathlib import Path

from git import Repo

GitInfo = namedtuple(
    "GitInfo",
    [
        "directory",
        "code_diff",
        "code_diff_staged",
        "commit_hash",
        "branch_name",
    ],
)


def copyanything(src, dst):
    """
    Copy a directory tree from `src` to `dst`, ignoring any files or directories named "lfs".

    :param src: The path to the source directory.
    :type src: str
    :param dst: The path to the destination directory.
    :type dst: str
    """
    try:
        shutil.copytree(src, dst, ignore=shutil.ignore_patterns("lfs"))
    except OSError as exc:  # python >2.5
        if exc.errno in (errno.ENOTDIR, errno.EINVAL):
            shutil.copy(src, dst)
        else:
            raise


def copy_git_repo(root: str, dest_path: str, exclude: list = None):
    """Given a destination path, copy the local root dir to the path.
    This function can be used to generate a snapshot of the repo so that the
    exactly same code status will be recovered when later playing a trained
    model or launching a grid-search job in the waiting queue.
    Args:
        root (str): the path to the repo
        dest_path (str): the path to generate a snapshot of repo
    """
    root_path = Path(root)
    dest_path = Path(dest_path)
    # Get output from git ls-files --others --exclude-standard, split into list
    files = subprocess.run(
        ["git", "ls-files", "--cached", "--others", "--exclude-standard"],
        cwd=root_path,
        capture_output=True,
        text=True,
    ).stdout.splitlines()

    # Create the destination directory if it doesn't exist
    dest_path.mkdir(parents=True, exist_ok=True)
    # Copy files to destination, preserving directory structure

    # remove excluded
    if exclude is not None:
        for e in exclude:
            if e in files:
                files.remove(e)

    for src_file in files:
        src_file = Path(src_file)
        dest_file = dest_path / src_file
        # Create necessary directories in destination
        dest_file.parent.mkdir(parents=True, exist_ok=True)

        src = root_path / src_file
        if not src.exists():
            continue  # ignore deleted files that were listed by git ls-files

        # Copy file from source to destination
        copyanything(root_path / src_file, dest_file)

    copyanything(root_path / ".git", dest_path / ".git")

    # Recursively process submodules
    root_repo = Repo(root)
    submodules = get_submodules(root_repo)
    for submodule in submodules:
        # Call copy_git_repo recursively on the submodule
        if any([e in submodule.working_tree_dir for e in exclude]):
            continue
        copy_git_repo(
            submodule.working_tree_dir,
            dest_path / Path(submodule.working_tree_dir).relative_to(root_path),
        )


def generate_snapshot(root: str, dest_path: str, exclude: list = None):
    """
    Generate a snapshot of a git repository and its submodules.

    :param root: The path to the root of the git repository.
    :type root: str
    :param dest_path: The path to the directory where the snapshot will be saved.
    :type dest_path: str
    :param exclude: A list of paths to exclude from the snapshot, defaults to None.
    :type exclude: list, optional
    """
    copy_git_repo(root, dest_path, exclude=exclude)
    dest_path = Path(dest_path)

    root_repo = Repo(root)
    submodules = get_submodules_recursively(root_repo)

    git_infos = [get_git_info(i.working_tree_dir) for i in [root_repo] + submodules]
    for directory, code_diff, code_diff_staged, commit_hash, branch_name in git_infos:
        directory_path = Path(directory)
        diff_file_name = (directory_path.absolute().as_posix()[1:] + ".patch").replace(
            "/", "-"
        )
        diff_staged_file_name = (
            directory_path.absolute().as_posix()[1:] + "_staged.patch"
        ).replace("/", "-")

        if code_diff is not None and len(code_diff) > 0:
            with open(dest_path / diff_file_name, "w") as f:
                f.write(code_diff + "\n")
        if code_diff_staged is not None and len(code_diff_staged) > 0:
            with open(dest_path / diff_staged_file_name, "w") as f:
                f.write(code_diff_staged + "\n")

        with open(dest_path / "git_infos.txt", "a") as f:
            f.write(f"directory: {directory}\n")
            f.write(f"git hash: {commit_hash}\n")
            f.write(f"git branch name: {branch_name}\n\n")


def get_submodules_recursively(root: Repo):
    """
    Recursively retrieves all submodules of a given repository.

    :param root: The root repository to retrieve submodules from.
    :type root: Repo
    :return: A list of all submodules of the root repository.
    :rtype: list[Repo]
    """
    ret = []
    for i in get_submodules(root):
        ret += [i] + get_submodules_recursively(i)
    return ret


def get_submodules(root: Repo) -> List[Repo]:
    """
    Returns a list of Repo objects representing the submodules of the given root repository.

    :param root: The root repository.
    :type root: Repo
    :return: A list of Repo objects representing the submodules of the given root repository.
    :rtype: List[Repo]
    """
    return [Repo(Path(root.working_tree_dir) / i.path) for i in root.submodules]


def get_git_info(dir):
    """
    Returns GitInfo object containing information about the git repository located at `dir`.

    :param dir: Path to the git repository.
    :type dir: str
    :return: GitInfo object containing information about the git repository.
    :rtype: GitInfo
    """
    repo = Repo(dir)
    try:
        branch_name = repo.active_branch.name
    except TypeError:
        branch_name = "[DETACHED]"
    return GitInfo(
        directory=dir,
        code_diff=repo.git.diff(None),
        code_diff_staged=repo.git.diff("--staged"),
        commit_hash=repo.head.commit.hexsha,
        branch_name=branch_name,
    )



import torch
import wandb
from rich import print


def log_wandb_distribution(key, samples, quantiles: list = None):
    """
    https://github.com/ezhang7423/wandb-histogram-over-time/tree/main
    Log the distribution of sample data points in Weights & Biases (wandb).

    This function computes the specified quantiles of the provided samples and logs them to wandb.
    If Weights & Biases run is not active, it prints the distribution to the console using rich print.
    By default, quantiles are calculated for the 0.0001, 0.01, 0.05, 0.10, and 0.25 levels, as well as their
    corresponding levels above the median (0.5).

    :param key: The label under which to log the quantiles.
    :type key: str
    :param samples: An array of sample data points.
    :type samples: torch.FloatTensor or torch.DoubleTensor
    :param quantiles: A list of quantiles to calculate for the sample distribution, defaults to None.
                      If None, uses the default set of quantiles that focus on the lower half of the distribution.
    :type quantiles: list of float, optional

    example usage:
    >>> import wandb
    >>> wandb.init()
    >>> gaussian_process = []

    >>> for i in range(1, 100):
    >>>     data = torch.randn(100) * math.log(i)
    >>>     gaussian_process.append(data)
    >>>     log_wandb_distribution("gaussian_process", data)

    """
    assert len(samples.shape) == 1

    if quantiles is None:
        quantiles = [0.0001, 0.01, 0.05, 0.10, 0.25]

    # all values less than 0.5
    assert all([q < 0.5 for q in quantiles])
    quantiles = quantiles + [0.5] + [1 - q for q in quantiles[::-1]]  # add median

    dist_quantiles = torch.quantile(samples, torch.tensor(quantiles).to(samples))

    d = {key + f"/{q}": dist_quantiles[i].item() for i, q in enumerate(quantiles)}

    if wandb.run is None:
        print(d)
    else:
        wandb.log(d)
    # not a good solution since this is not editable over time
    # d = {
    #     key: wandb.plot.line_series(
    #         xs=list(range(dist_quantiles.shape[1])),
    #         ys=dist_quantiles,
    #         keys=quantiles,
    #         title="Distribution over time",
    #         xname="Time Step",
    #     )
    # }



import collections

"""
Dictionary methods
"""


def nested_dict_to_dot_map_dict(d, parent_key=""):
    """
    Convert a recursive dictionary into a flat, dot-map dictionary.

    :param d: e.g. {'a': {'b': 2, 'c': 3}}
    :param parent_key: Used for recursion
    :return: e.g. {'a.b': 2, 'a.c': 3}
    """
    items = []
    for k, v in d.items():
        new_key = parent_key + "." + k if parent_key else k
        if isinstance(v, collections.abc.MutableMapping):
            items.extend(nested_dict_to_dot_map_dict(v, new_key).items())
        else:
            items.append((new_key, v))
    return dict(items)


def merge_recursive_dicts(a, b, path=None, ignore_duplicate_keys_in_second_dict=False):
    """
    Merge two dictionaries that may have nested dictionaries.

    :param a: The first dictionary to merge.
    :type a: dict
    :param b: The second dictionary to merge.
    :type b: dict
    :param path: The path to the current key being merged, used for error reporting. Defaults to None.
    :type path: list, optional
    :param ignore_duplicate_keys_in_second_dict: Whether to ignore duplicate keys in the second dictionary. Defaults to False.
    :type ignore_duplicate_keys_in_second_dict: bool, optional
    :raises Exception: If there are duplicate keys and ignore_duplicate_keys_in_second_dict is False.
    :return: The merged dictionary.
    :rtype: dict
    """

    if path is None:
        path = []
    for key in b:
        if key in a:
            if isinstance(a[key], dict) and isinstance(b[key], dict):
                merge_recursive_dicts(
                    a[key],
                    b[key],
                    path + [str(key)],
                    ignore_duplicate_keys_in_second_dict=ignore_duplicate_keys_in_second_dict,
                )
            elif a[key] == b[key]:
                print(f"Same value for key: {key}")
            else:
                duplicate_key = ".".join(path + [str(key)])
                if ignore_duplicate_keys_in_second_dict:
                    print(f"duplicate key ignored: {duplicate_key}")
                else:
                    raise Exception(f"Duplicate keys at {duplicate_key}")
        else:
            a[key] = b[key]
    return a


def list_of_dicts__to__dict_of_lists(lst):
    """
    Convert a list of dictionaries into a dictionary of lists, where each key in the output dictionary
    corresponds to a key in the input dictionaries, and each value in the output dictionary is a list
    containing the values of that key in the input dictionaries.

    :param lst: A list of dictionaries.
    :type lst: list[dict]

    :return: A dictionary of lists.
    :rtype: dict

    Example:
    --------
    >>> x = [
    ...     {'foo': 3, 'bar': 1},
    ...     {'foo': 4, 'bar': 2},
    ...     {'foo': 5, 'bar': 3},
    ... ]
    >>> list_of_dicts__to__dict_of_lists(x)
    {'foo': [3, 4, 5], 'bar': [1, 2, 3]}
    """
    if len(lst) == 0:
        return {}
    keys = lst[0].keys()
    output_dict = collections.defaultdict(list)
    for d in lst:
        assert set(d.keys()) == set(keys), (d.keys(), keys)
        for k in keys:
            output_dict[k].append(d[k])
    return output_dict


def safe_json(data):
    """
    Check if a given data structure is JSON serializable.

    :param data: The data structure to check.
    :type data: Any
    :return: True if the data structure is JSON serializable, False otherwise.
    :rtype: bool
    """
    if data is None:
        return True
    elif isinstance(data, (bool, int, float)):
        return True
    elif isinstance(data, (tuple, list)):
        return all(safe_json(x) for x in data)
    elif isinstance(data, dict):
        return all(isinstance(k, str) and safe_json(v) for k, v in data.items())
    return False


def dict_to_safe_json(d, sort=False):
    """
    Convert each value in the dictionary into a JSON'able primitive.

    :param d: The dictionary to convert.
    :type d: dict or collections.OrderedDict
    :param sort: Whether to sort the resulting dictionary by key. Defaults to False.
    :type sort: bool, optional
    :return: A new dictionary with all values converted to JSON'able primitives.
    :rtype: dict or collections.OrderedDict
    """
    if isinstance(d, collections.OrderedDict):
        new_d = collections.OrderedDict()
    else:
        new_d = {}
    for key, item in d.items():
        if safe_json(item):
            new_d[key] = item
        else:
            if isinstance(item, dict) or isinstance(item, collections.OrderedDict):
                new_d[key] = dict_to_safe_json(item, sort=sort)
            else:
                new_d[key] = str(item)
    if sort:
        return collections.OrderedDict(sorted(new_d.items()))
    else:
        return new_d


def recursive_items(dictionary):
    """
    Get all (key, item) recursively in a potentially recursive dictionary.
    Usage:

    ```
    x = {
        'foo' : {
            'bar' : 5
        }
    }
    recursive_items(x)
    # output:
    # ('foo', {'bar' : 5})
    # ('bar', 5)
    ```
    :param dictionary:
    :return:
    """
    for key, value in dictionary.items():
        yield key, value
        if isinstance(value, dict):
            yield from recursive_items(value)



def register_buffer(self, name, val):
    """Registers a buffer as the same type as weights for lightning modules.

    :param name: The name of the buffer.
    :type name: str
    :param val: The value of the buffer.
    :type val: torch.Tensor
    """
    self.register_buffer(name, val.type(self.dtype))



"""
GPU wrappers
"""
import torch

from eztils import default


def device_dtype_decorator(func):
    """
    A decorator function that adds device and dtype parameters to a given function.

    :param func: The function to be decorated.
    :type func: function
    :return: The decorated function.
    :rtype: function
    """

    def wrapper(*args, **kwargs):
        from eztils.torch import DEVICE, DTYPE

        torch_device = default(kwargs.get("device"), DEVICE)
        torch_dtype = default(kwargs.get("dtype"), DTYPE)
        kwargs.pop("device", None)
        kwargs.pop("dtype", None)
        return func(*args, **kwargs, device=torch_device, dtype=torch_dtype)

    return wrapper


# torch.from_numpy doesn't support kwargs
def from_numpy(np_array, device=None, dtype=None):
    """
    Create a PyTorch tensor from a numpy array.

    :param np_array: The numpy array to be converted to a PyTorch tensor.
    :type np_array: numpy.ndarray
    :param device: The device on which the tensor will be created. If None, the default device will be used.
    :type device: torch.device, optional
    :param dtype: The data type of the tensor. If None, the default data type will be used.
    :type dtype: torch.dtype, optional
    :return: A PyTorch tensor created from the input numpy array.
    :rtype: torch.Tensor
    """
    from eztils.torch import DEVICE, DTYPE

    torch_device = default(device, DEVICE)
    torch_dtype = default(dtype, DTYPE)

    return torch.from_numpy(np_array).to(device=torch_device, dtype=torch_dtype)


@device_dtype_decorator
def tensor(*args, **kwargs):
    return torch.tensor(*args, **kwargs)


@device_dtype_decorator
def randint(*sizes, **kwargs):
    return torch.randint(*sizes, **kwargs)


@device_dtype_decorator
def zeros(*sizes, **kwargs):
    return torch.zeros(*sizes, **kwargs)


@device_dtype_decorator
def empty(*sizes, **kwargs):
    return torch.empty(*sizes, **kwargs)


@device_dtype_decorator
def ones(*sizes, **kwargs):
    return torch.ones(*sizes, **kwargs)


@device_dtype_decorator
def ones_like(*args, **kwargs):
    return torch.ones_like(*args, **kwargs)


@device_dtype_decorator
def randn(*args, **kwargs):
    return torch.randn(*args, **kwargs)


@device_dtype_decorator
def zeros_like(*args, **kwargs):
    return torch.zeros_like(*args, **kwargs)


@device_dtype_decorator
def randperm(*args, **kwargs):
    return torch.randperm(*args, **kwargs)


@device_dtype_decorator
def rand(*args, **kwargs):
    return torch.rand(*args, **kwargs)


@device_dtype_decorator
def normal(*args, **kwargs):
    return torch.normal(*args, **kwargs)



"""
import often used modules
"""
import random
from functools import partial

import numpy as np
import torch

from eztils import load, save

"""
globals
"""
USE_GPU = False
DTYPE = torch.float
GPU_ID = 0
DEVICE = None


def set_gpu_mode(mode, gpu_id=0):
    """
    Sets the mode and GPU ID for PyTorch.

    :param mode: A boolean indicating whether to use GPU or not.
    :type mode: bool
    :param gpu_id: The ID of the GPU to use, defaults to 0.
    :type gpu_id: int, optional
    """
    global USE_GPU
    global DEVICE
    global GPU_ID
    GPU_ID = gpu_id
    USE_GPU = mode
    DEVICE = torch.device("cuda:" + str(gpu_id) if USE_GPU else "cpu")
    if USE_GPU:
        torch.cuda.set_device(gpu_id)


def activation_from_string(string):
    """
    Returns the activation function corresponding to the given string.

    :param string: The name of the activation function to return.
    :type string: str
    :return: The activation function.
    :rtype: function
    """
    if string == "identity":
        return lambda x: x
    return getattr(nn, string)()


def soft_update_from_to(source, target, tau):
    """
    Update the target model parameters with a fraction of the source model parameters.

    :param source: The source model whose parameters will be used to update the target model.
    :type source: torch.nn.Module
    :param target: The target model whose parameters will be updated.
    :type target: torch.nn.Module
    :param tau: The fraction of the source model parameters to be used for updating the target model parameters.
    :type tau: float
    """
    for target_param, param in zip(target.parameters(), source.parameters()):
        target_param.data.copy_(target_param.data * (1.0 - tau) + param.data * tau)


"""
helper functions
"""


def seed_everything(seed):
    """
    Set the seed for all the possible random number generators
    for global packages.
    :param seed:
    :return: None
    """
    seed = int(seed)
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)

    if torch.cuda.is_available():
        torch.cuda.manual_seed_all(seed)


save = partial(save, save_fn=torch.save)
torch.load = partial(torch.load, map_location="cpu")
load = partial(load, load_fn=torch.load)


def get_best_cuda() -> int:
    import numpy as np
    import pynvml

    pynvml.nvmlInit()
    deviceCount = pynvml.nvmlDeviceGetCount()
    deviceMemory = []
    for i in range(deviceCount):
        handle = pynvml.nvmlDeviceGetHandleByIndex(i)
        mem_info = pynvml.nvmlDeviceGetMemoryInfo(handle)
        deviceMemory.append(mem_info.free)
    deviceMemory = np.array(deviceMemory, dtype=np.int64)
    best_device_index = np.argmax(deviceMemory)
    print("best gpu:", best_device_index)
    return best_device_index.item()


def freeze(model: torch.nn.Module):
    for p in model.parameters():
        p.requires_grad = True


def identity(x):
    return x


from .distributions import *
from .lightning import *
from .math import *
from .modules import *
from .parameters import *
from .tensor_creators import *
from .to import *



import os
import pickle
import re
from multiprocessing import shared_memory
from pathlib import Path

import loguru
from varname import ImproperUseError, argname


def remove_shm_from_resource_tracker():
    """
    Monkey patch multiprocessing.resource_tracker so SharedMemory won't be tracked
    More details at: https://bugs.python.org/issue38119
    """
    # pylint: disable=protected-access, import-outside-toplevel
    # Ignore linting errors in this bug workaround hack
    from multiprocessing import resource_tracker

    def fix_register(name, rtype):
        if rtype == "shared_memory":
            return None
        return resource_tracker._resource_tracker.register(name, rtype)

    resource_tracker.register = fix_register

    def fix_unregister(name, rtype):
        if rtype == "shared_memory":
            return None
        return resource_tracker._resource_tracker.unregister(name, rtype)

    resource_tracker.unregister = fix_unregister
    if "shared_memory" in resource_tracker._CLEANUP_FUNCS:
        del resource_tracker._CLEANUP_FUNCS["shared_memory"]


# More details at: https://bugs.python.org/issue38119
remove_shm_from_resource_tracker()


def make_valid_python_var(varStr):
    """
    Replaces all non-alphanumeric characters in a string with underscores, and ensures that the resulting string is a valid Python variable name.

    :param varStr: The string to convert to a valid Python variable name.
    :type varStr: str
    :return: The input string with non-alphanumeric characters replaced with underscores, and with any leading digits replaced with an underscore.
    :rtype: str
    """
    return re.sub(r"\W|^(?=\d)", "_", varStr)


def save(var, folder=None, memory=False, save_fn=pickle.dump):
    """
    Save a variable to disk or shared memory.

    :param var: The variable to be saved.
    :type var: Any
    :param folder: The folder where the variable should be saved. Defaults to current directory.
    :type folder: str, optional
    :param memory: Whether to save the variable to shared memory. Defaults to False.
    :type memory: bool, optional
    :param save_fn: The function to use for saving the variable. Defaults to pickle.dump.
    :type save_fn: callable, optional
    """
    folder = Path(folder or ".").absolute().resolve()
    try:
        varname = make_valid_python_var(argname("var"))
    except ImproperUseError:
        varname = "tensor"

    if not memory:
        fname = str(folder / f"{varname}.pt")
        loguru.logger.info(f"Saving {varname} to {fname}")
        save_fn(var, fname)
    else:
        script = f"""
        from multiprocessing import shared_memory
        handle = shared_memory.SharedMemory(name=\\"{varname}\\", create=False)
        breakpoint()
        """.strip().replace(
            "\n", ";"
        )
        var_bytes = pickle.dumps(var)
        try:
            # see if exists
            s = shared_memory.SharedMemory(name=varname)
            s.close()
            s.unlink()
        except FileNotFoundError:
            pass

        shm = shared_memory.SharedMemory(create=True, size=len(var_bytes), name=varname)

        shm.buf[:] = var_bytes
        loguru.logger.info(f"Saved {varname} to shared memory at /dev/shm/{varname}")

        # so that the shared memory is not deleted when the script exits, as a handle still exits
        os.system(f"""tmux new-session -d -s {varname}""")
        os.system(f"""tmux send-keys -t {varname} "python -c '{script}'" Enter""")


def load(path, memory=False, load_fn=pickle.load, **kwargs):
    """
    Load a serialized object from a file or shared memory.

    :param path: The path to the file or the name of the shared memory.
    :type path: str
    :param memory: If True, load from shared memory. Otherwise, load from file. Defaults to False.
    :type memory: bool, optional
    :param load_fn: The function to use for loading the serialized object. Defaults to pickle.load.
    :type load_fn: callable, optional
    :param kwargs: Additional keyword arguments to pass to the load function.
    :return: The deserialized object.
    :rtype: Any
    """
    if not memory:
        return load_fn(path, **kwargs)
    else:
        shm = shared_memory.SharedMemory(name=path, create=False)
        return pickle.loads(shm.buf)



from collections import OrderedDict
from numbers import Number

import numpy as np


def normalize(x):
    """
    Normalize an array by scaling its values between 0 and 1.

    :param x: The input array to be normalized.
    :type x: numpy.ndarray
    :return: The normalized array.
    :rtype: numpy.ndarray
    """
    x = x - x.min()
    x = x / x.max()
    return x


def create_stats_ordered_dict(
    name,
    data,
    stat_prefix=None,
    always_show_all_stats=True,
    exclude_max_min=False,
):
    """
    Create an ordered dictionary of statistics for the given data.

    :param name: The name of the data.
    :type name: str
    :param data: The data to compute statistics for.
    :type data: Union[Number, List[Number], Tuple[Number], np.ndarray]
    :param stat_prefix: A prefix to add to the name of the data, defaults to None.
    :type stat_prefix: Optional[str]
    :param always_show_all_stats: Whether to always show all statistics, defaults to True.
    :type always_show_all_stats: bool, optional
    :param exclude_max_min: Whether to exclude the maximum and minimum values, defaults to False.
    :type exclude_max_min: bool, optional
    :return: An ordered dictionary of statistics for the given data.
    :rtype: OrderedDict
    """
    if stat_prefix is not None:
        name = f"{stat_prefix}{name}"
    if isinstance(data, Number):
        return OrderedDict({name: data})

    if len(data) == 0:
        return OrderedDict()

    if isinstance(data, tuple):
        ordered_dict = OrderedDict()
        for number, d in enumerate(data):
            sub_dict = create_stats_ordered_dict(
                f"{name}_{number}",
                d,
            )
            ordered_dict.update(sub_dict)
        return ordered_dict

    if isinstance(data, list):
        try:
            iter(data[0])
        except TypeError:
            pass
        else:
            data = np.concatenate(data)

    if isinstance(data, np.ndarray) and data.size == 1 and not always_show_all_stats:
        return OrderedDict({name: float(data)})

    stats = OrderedDict(
        [
            (name + " Mean", np.mean(data)),
            (name + " Std", np.std(data)),
        ]
    )
    if not exclude_max_min:
        stats[name + " Max"] = np.max(data)
        stats[name + " Min"] = np.min(data)
    return stats



"""
Add custom distributions in addition to th existing ones
"""
from collections import OrderedDict

import torch
import torch.nn.functional as F
from torch.distributions import Bernoulli as TorchBernoulli
from torch.distributions import Beta as TorchBeta
from torch.distributions import Distribution as TorchDistribution
from torch.distributions import Independent as TorchIndependent
from torch.distributions import Normal as TorchNormal
from torch.distributions import kl_divergence
from torch.distributions.utils import _sum_rightmost

from eztils.default.ezmath import create_stats_ordered_dict
from eztils.torch.math import atanh
from eztils.torch.tensor_creators import ones, tensor, zeros
from eztils.torch.to import to_np


class Distribution(TorchDistribution):
    def sample_and_logprob(self):
        s = self.sample()
        log_p = self.log_prob(s)
        return s, log_p

    def rsample_and_logprob(self):
        s = self.rsample()
        log_p = self.log_prob(s)
        return s, log_p

    def mle_estimate(self):
        return self.mean

    def get_diagnostics(self):
        return {}


class TorchDistributionWrapper(Distribution):
    def __init__(self, distribution: TorchDistribution):
        self.distribution = distribution

    @property
    def batch_shape(self):
        return self.distribution.batch_shape

    @property
    def event_shape(self):
        return self.distribution.event_shape

    @property
    def arg_constraints(self):
        return self.distribution.arg_constraints

    @property
    def support(self):
        return self.distribution.support

    @property
    def mean(self):
        return self.distribution.mean

    @property
    def variance(self):
        return self.distribution.variance

    @property
    def stddev(self):
        return self.distribution.stddev

    def sample(self, sample_size=torch.Size()):
        return self.distribution.sample(sample_shape=sample_size)

    def rsample(self, sample_size=torch.Size()):
        return self.distribution.rsample(sample_shape=sample_size)

    def log_prob(self, value):
        return self.distribution.log_prob(value)

    def cdf(self, value):
        return self.distribution.cdf(value)

    def icdf(self, value):
        return self.distribution.icdf(value)

    def enumerate_support(self, expand=True):
        return self.distribution.enumerate_support(expand=expand)

    def entropy(self):
        return self.distribution.entropy()

    def perplexity(self):
        return self.distribution.perplexity()

    def __repr__(self):
        return "Wrapped " + self.distribution.__repr__()


class Delta(Distribution):
    """A deterministic distribution"""

    def __init__(self, value):
        self.value = value

    def sample(self):
        return self.value.detach()

    def rsample(self):
        return self.value

    @property
    def mean(self):
        return self.value

    @property
    def variance(self):
        return 0

    @property
    def entropy(self):
        return 0


class TanhDelta(Delta):
    def __init__(self, value):
        super().__init__(torch.tanh(value))
        self.pre_tanh_value = value


class Bernoulli(Distribution, TorchBernoulli):
    def get_diagnostics(self):
        stats = OrderedDict()
        stats.update(
            create_stats_ordered_dict(
                "probability",
                to_np(self.probs),
            )
        )
        return stats


class Independent(Distribution, TorchIndependent):
    def get_diagnostics(self):
        return self.base_dist.get_diagnostics()


class Beta(Distribution, TorchBeta):
    def get_diagnostics(self):
        stats = OrderedDict()
        stats.update(
            create_stats_ordered_dict(
                "alpha",
                to_np(self.concentration0),
            )
        )
        stats.update(
            create_stats_ordered_dict(
                "beta",
                to_np(self.concentration1),
            )
        )
        stats.update(
            create_stats_ordered_dict(
                "entropy",
                to_np(self.entropy()),
            )
        )
        return stats


class MultivariateDiagonalNormal(TorchDistributionWrapper):
    from torch.distributions import constraints

    arg_constraints = {"loc": constraints.real, "scale": constraints.positive}

    def __init__(self, loc, scale_diag, reinterpreted_batch_ndims=1):
        dist = Independent(
            TorchNormal(loc, scale_diag),
            reinterpreted_batch_ndims=reinterpreted_batch_ndims,
        )
        super().__init__(dist)

    def get_diagnostics(self):
        stats = OrderedDict()
        stats.update(
            create_stats_ordered_dict(
                "mean",
                to_np(self.mean),
                # exclude_max_min=True,
            )
        )
        stats.update(
            create_stats_ordered_dict(
                "std",
                to_np(self.distribution.stddev),
            )
        )
        return stats

    def __repr__(self):
        return self.distribution.base_dist.__repr__()


# Independent RV KL handling - https://github.com/pytorch/pytorch/issues/13545


@torch.distributions.kl.register_kl(TorchIndependent, TorchIndependent)
def _kl_independent_independent(p, q):
    if p.reinterpreted_batch_ndims != q.reinterpreted_batch_ndims:
        raise NotImplementedError
    result = kl_divergence(p.base_dist, q.base_dist)
    return _sum_rightmost(result, p.reinterpreted_batch_ndims)


class TanhGaussianMixture(TorchDistributionWrapper):
    """
    Construct a gaussian mixture model using gumbel softmax for backprop.
    https://bochang.me/blog/posts/pytorch-distributions/#table

    """

    def __init__(
        self,
        mean,  #: TensorType["batch", "num_gaussians", "action_dim"],
        std,  #: TensorType["batch", "num_gaussians", "action_dim"],
        weights,  #: TensorType["batch", "num_gaussians"],
    ):
        if len(mean.shape) == 2:
            mean = mean.unsqueeze(dim=0)
        if len(std.shape) == 2:
            std = std.unsqueeze(dim=0)
        if len(weights.shape) == 1:
            weights = weights.unsqueeze(dim=0)

        weights_sum = weights.sum(dim=-1)
        assert torch.all(
            torch.isclose(weights_sum, torch.ones_like(weights_sum))
        ), "Weights must sum to 1"

        self.num_gaussians = weights.shape[-1]
        self.batch_size = weights.shape[0]
        self.normal_mean = mean
        self.weights = weights
        super().__init__(MultivariateDiagonalNormal(mean, std))

    def mle_estimate(self, return_std=False):
        """Return the mean of the most likely component.
        This often computes the mode of the distribution, but not always.
        """
        bz = torch.arange(self.batch_size)
        ind = torch.argmax(self.weights, dim=1)

        if not return_std:
            return self.mean[bz, ind]
        else:
            return self.mean[bz, ind], self.stddev[bz, ind]

    def sample(self, n=1):
        assert isinstance(n, int), f"{n} must be an int"
        return self.rsample(n=n, hard=True)

    def rsample(self, n=1, temperature=0.02, hard=False):
        """Differentiable sample with gumbel softmax

        Returns
        ----------
        tensor of size (n, num_feat)
        """
        assert isinstance(n, int), f"{n} must be an int"

        soft_categorical = F.gumbel_softmax(
            self.weights.log().unsqueeze(0).expand(n, -1, -1),
            tau=temperature,
            hard=hard,
        )
        soft_categorical = soft_categorical.unsqueeze(-1)
        sample = self.distribution.rsample([n])  # [n, batch, num_gaussian, dim]
        return torch.tanh(torch.einsum("hijk,hijk->hik", sample, soft_categorical))

    def _log_prob_from_pre_tanh(self, pre_tanh_value):
        if (
            len(pre_tanh_value.shape) == 3
        ):  # [n, batch, action_dim] -> Require [1, batch, action_dim]
            # this came from calling dist.sample()
            # assert pre_tanh_value.shape[0] == 1
            assert pre_tanh_value.shape[1] == self.batch_size
            if pre_tanh_value.shape[0] == 1:
                pre_tanh_value = pre_tanh_value.squeeze(0)

        # [batch_size, action_dim] -> [batch_size, num_gaussians, action_dim]
        pre_tanh_value_mg = pre_tanh_value.unsqueeze(-2)
        # [batch_size, num_gaussians]
        # equivalent to the jacobian, -log(1 - tanh(x)^2).
        correction = -2.0 * (
            tensor([2.0]).log() - pre_tanh_value - F.softplus(-2.0 * pre_tanh_value)
        ).sum(dim=-1)
        # [batch_size, num_gaussians]
        pi = torch.exp(self.distribution.log_prob(pre_tanh_value_mg)).clip(1e-40)
        ret = torch.log((self.weights * pi).sum(-1)) + correction
        return ret

    def log_prob(self, value, pre_tanh_value=None):
        assert (
            value.shape[-1] == self.mean.shape[-1]
        ), f"{value} must have shape [{self.batch_shape[0]}, {self.batch_shape[1]}, {self.event_shape[0]}]"

        if pre_tanh_value is None:
            # errors or instability at values near 1
            pre_tanh_value = atanh(value)
        else:
            assert pre_tanh_value.shape[-1] == self.mean.shape[-1]

        return self._log_prob_from_pre_tanh(pre_tanh_value)

    def __repr__(self):
        s = "TanhGaussianMixture(mean=%s, std=%s, weights=%s)"
        return s % (self.mean, self.stddev, self.weights)

    @property
    def mean(self):
        return torch.tanh(self.normal_mean)

    def perplexity(self):
        return NotImplementedError

    def entropy(self):
        return NotImplementedError

    def get_diagnostics(self):
        stats = OrderedDict()
        stats.update(
            create_stats_ordered_dict(
                "mean",
                to_np(self.mean),
            )
        )
        stats.update(
            create_stats_ordered_dict(
                "stddev",
                to_np(self.stddev),
            )
        )
        stats.update(
            create_stats_ordered_dict(
                "weights",
                to_np(self.weights),
            )
        )
        return stats


class GaussianMixture(TanhGaussianMixture):
    def __init__(
        self,
        mean,  #: TensorType["batch", "num_gaussians", "action_dim"],
        std,  #: TensorType["batch", "num_gaussians", "action_dim"],
        weights,  #: TensorType["batch", "num_gaussians"],
    ):
        super().__init__(mean, std, weights)

    def log_prob(self, value):
        assert (
            value.shape[-1] == self.mean.shape[-1]
        ), f"{value} must have shape [{self.batch_shape[0]}, {self.batch_shape[1]}, {self.event_shape[0]}]"
        if (
            len(value.shape) == 3
        ):  # [n, batch, action_dim] -> Require [1, batch, action_dim]
            # this came from calling dist.sample()
            assert value.shape[0] == 1
            assert value.shape[1] == self.batch_size
            value = value.squeeze(0)
        value = value.unsqueeze(1).expand(-1, self.num_gaussians, -1)
        pi = torch.exp(self.distribution.log_prob(value)).clip(1e-40)
        return torch.log(torch.einsum("ij,ij->i", self.weights, pi))

    def rsample(self, n=1, temperature=0.02, hard=False):
        return torch.atanh(super().rsample(n, temperature, hard))

    def __repr__(self):
        s = "GaussianMixture(mean=%s, std=%s, weights=%s)"
        return s % (self.mean, self.stddev, self.weights)

    @property
    def mean(self):
        return self.normal_mean


class TanhNormal(Distribution):
    """
    Represent distribution of X where
        X ~ tanh(Z)
        Z ~ N(mean, std)

    Note: this is not very numerically stable.
    """

    def __init__(self, normal_mean, normal_std, epsilon=1e-6):
        """
        :param normal_mean: Mean of the normal distribution
        :param normal_std: Std of the normal distribution
        :param epsilon: Numerical stability epsilon when computing log-prob.
        """
        self.normal_mean = normal_mean
        self.normal_std = normal_std
        self.normal = MultivariateDiagonalNormal(normal_mean, normal_std)
        self.epsilon = epsilon

    def sample_n(self, n, return_pre_tanh_value=False):
        z = self.normal.sample_n(n)
        if return_pre_tanh_value:
            return torch.tanh(z), z
        else:
            return torch.tanh(z)

    def _log_prob_from_pre_tanh(self, pre_tanh_value):
        """
        Adapted from
        https://github.com/tensorflow/probability/blob/master/tensorflow_probability/python/bijectors/tanh.py#L73

        This formula is mathematically equivalent to log(1 - tanh(x)^2).

        Derivation:

        log(1 - tanh(x)^2)
         = log(sech(x)^2)
         = 2 * log(sech(x))
         = 2 * log(2e^-x / (e^-2x + 1))
         = 2 * (log(2) - x - log(e^-2x + 1))
         = 2 * (log(2) - x - softplus(-2x))

        :param value: some value, x
        :param pre_tanh_value: arctanh(x)
        :return:
        """
        log_prob = self.normal.log_prob(pre_tanh_value)
        correction = -2.0 * (
            tensor([2.0]).log()
            - pre_tanh_value
            - torch.nn.functional.softplus(-2.0 * pre_tanh_value)
        ).sum(dim=-1)
        return log_prob + correction

    def log_prob(self, value, pre_tanh_value=None):
        if pre_tanh_value is None:
            # errors or instability at values near 1
            value = torch.clamp(value, -0.999999, 0.999999)
            pre_tanh_value = torch.log(1 + value) / 2 - torch.log(1 - value) / 2
        return self._log_prob_from_pre_tanh(pre_tanh_value)

    def rsample_with_pretanh(self, sample_shape=torch.Size()):
        z = self.normal_mean + self.normal_std * MultivariateDiagonalNormal(
            zeros(self.normal_mean.size()), ones(self.normal_std.size())
        ).sample(sample_shape)
        return torch.tanh(z), z

    def sample(self):
        """
        Gradients will and should *not* pass through this operation.

        See https://github.com/pytorch/pytorch/issues/4620 for discussion.
        """
        value, pre_tanh_value = self.rsample_with_pretanh()
        return value.detach()

    def rsample(self, sample_shape=torch.Size()):
        """
        Sampling in the reparameterization case.
        """
        value, pre_tanh_value = self.rsample_with_pretanh(sample_shape)
        return value

    def sample_and_logprob(self):
        value, pre_tanh_value = self.rsample_with_pretanh()
        value, pre_tanh_value = value.detach(), pre_tanh_value.detach()
        log_p = self.log_prob(value, pre_tanh_value)
        return value, log_p

    def rsample_and_logprob(self, sample_shape=torch.Size()):
        value, pre_tanh_value = self.rsample_with_pretanh(sample_shape)
        log_p = self.log_prob(value, pre_tanh_value)
        return value, log_p

    @property
    def mean(self):
        return torch.tanh(self.normal_mean)

    @property
    def stddev(self):
        return self.normal_std

    def __repr__(self):
        return f"TanhNormal (shape: {self.normal_mean.shape})"

    def get_diagnostics(self):
        stats = OrderedDict()
        stats.update(
            create_stats_ordered_dict(
                "mean",
                to_np(self.mean),
            )
        )
        stats.update(create_stats_ordered_dict("normal/std", to_np(self.normal_std)))
        stats.update(
            create_stats_ordered_dict(
                "normal/log_std",
                to_np(torch.log(self.normal_std)),
            )
        )
        return stats



"""eds utilities"""

import dataclasses
import functools
import inspect as inspect_
import json
import os
import sys
from dataclasses import dataclass
from datetime import datetime
from inspect import getsourcefile, isfunction
from pathlib import Path

from beartype import beartype

# import often used modules

"""
Get and set version of module
"""

if sys.version_info >= (3, 8):
    from importlib import metadata as importlib_metadata
else:
    import importlib_metadata


""""""


class EnhancedJSONEncoder(json.JSONEncoder):
    def default(self, o):
        if dataclasses.is_dataclass(o):
            return dataclasses.asdict(o)
        return super().default(o)


json.dumps = functools.partial(
    json.dumps, cls=EnhancedJSONEncoder
)  # monkey patch json.dumps to support dataclasses
json.dump = functools.partial(json.dump, cls=EnhancedJSONEncoder)
""""""


def get_version() -> str:
    try:
        return importlib_metadata.version(__name__)
    except importlib_metadata.PackageNotFoundError:  # pragma: no cover
        return "unknown"


version: str = get_version()
__version__ = version


"""
miscellaneous functions that are often used
"""


def default(val, d):
    """
    If val exists, return it. Otherwise, return d

    :param val: The value to check
    :param d: The default value to return if val is None
    :return: val if it exists, otherwise d
    """
    if val is not None:
        return val
    return d() if isfunction(d) else d


def cycle(dl):
    while True:
        yield from dl


def abspath():
    """
    Returns the absolute path of the directory containing the caller module.

    :return: The absolute path of the directory containing the caller module.
    :rtype: str

    # https://stackoverflow.com/questions/16771894/python-nameerror-global-name-file-is-not-defined
    # https://docs.python.org/3/library/inspect.html#inspect.FrameInfo
    # return os.path.dirname(inspect.stack()[1][1]) # type: ignore
    # return os.path.dirname(getsourcefile(lambda:0)) # type: ignore
    """
    return Path(os.path.dirname(getsourcefile(inspect_.stack()[1][0])))  # type: ignore


from pathlib import Path


def setup_path(path: str) -> Path:
    """
    Create path if it does not exist.

    :param path: The path to be created.
    :type path: str
    :return: The created path.
    :rtype: Path
    """
    path = Path(path).resolve()
    path.mkdir(parents=True, exist_ok=True)
    return path


def apply_dict(fn, d, *args, **kwargs):
    """
    Apply a function to each value in a dictionary and return a new dictionary with the results.

    :param fn: The function to apply to each value in the dictionary.
    :type fn: function
    :param d: The dictionary to apply the function to.
    :type d: dict
    :param args: Additional positional arguments to pass to the function.
    :type args: tuple
    :param kwargs: Additional keyword arguments to pass to the function.
    :type kwargs: dict
    :return: A new dictionary with the results of applying the function to each value in the input dictionary.
    :rtype: dict
    """
    return {k: fn(v, *args, **kwargs) for k, v in d.items()}


def query_yes_no(question, default="yes"):
    """Ask a yes/no question via raw_input() and return their answer.

    "question" is a string that is presented to the user.
    "default" is the presumed answer if the user just hits <Enter>.
        It must be "yes" (the default), "no" or None (meaning
        an answer is required of the user).

    The "answer" return value is True for "yes" or False for "no".
    """
    valid = {"yes": True, "y": True, "ye": True, "no": False, "n": False}
    if default is None:
        prompt = " [y/n] "
    elif default == "yes":
        prompt = " [Y/n] "
    elif default == "no":
        prompt = " [y/N] "
    else:
        raise ValueError("invalid default answer: '%s'" % default)

    while True:
        sys.stdout.write(question + prompt)
        choice = input().lower()
        if default is not None and choice == "":
            return valid[default]
        elif choice in valid:
            return valid[choice]
        else:
            sys.stdout.write("Please respond with 'yes' or 'no' " "(or 'y' or 'n').\n")


def datestr(full=True):
    """
    Returns a formatted date string.

    :param full: If True, includes seconds in the time. If False, only includes hours and minutes. Defaults to True.
    :type full: bool, optional
    :return: A string representing the current date and time in the format 'YYYY-MM-DD---HH-MM-SS' or 'YYYY-MM-DD---HH-MM', depending on the value of `full`.
    :rtype: str
    """
    now = datetime.now()
    if full:
        return f'{now.strftime("%Y-%m-%d")}---{now.strftime("%H-%M-%S")}'
    else:
        return f'{now.strftime("%Y-%m-%d")}---{now.strftime("%H-%M")}'


def wlog(*args, **kwargs):
    import wandb

    if wandb.run is not None:
        wandb.log(*args, **kwargs)


"""
often used modules
"""
# beartype

enforced_dataclass = beartype(dataclass)
frozen_enforced_dataclass = beartype(dataclass(frozen=True))



# -----------------------------------------------------------------------------#
# ----------------------------- parameter counting ----------------------------#
# -----------------------------------------------------------------------------#
# https://github.com/allenai/allenact/blob/f00445e4ae8724ccc001b3300af5c56a7f882614/allenact/utils/tensor_utils.py#L1


def param_to_module(param):
    """
    Extracts the module name from a parameter string.

    :param param: A string representing a parameter in the format "module.submodule.parameter".
    :type param: str
    :return: The name of the module containing the parameter.
    :rtype: str
    """
    module_name = param[::-1].split(".", maxsplit=1)[-1][::-1]
    return module_name


def report_parameters(model, topk=10):
    """
    Prints a summary of the total number of parameters in a PyTorch model, as well as the top-k parameters with the largest number of elements.

    :param model: The PyTorch model to analyze.
    :type model: torch.nn.Module
    :param topk: The number of parameters to display in the summary. Defaults to 10.
    :type topk: int, optional
    :return: The total number of parameters in the model.
    :rtype: int
    """

    def _to_str(num):
        if num >= 1e6:
            return f"{(num/1e6):.2f} M"
        else:
            return f"{(num/1e3):.2f} k"

    counts = {k: p.numel() for k, p in model.named_parameters()}
    n_parameters = sum(counts.values())
    print(f"[ utils/arrays ] Total parameters: {_to_str(n_parameters)}")

    modules = dict(model.named_modules())
    sorted_keys = sorted(counts, key=lambda x: -counts[x])
    # max_length = max([len(k) for k in sorted_keys])
    for i in range(topk):
        key = sorted_keys[i]
        count = counts[key]
        module = param_to_module(key)
        print(" " * 8, f"{key:10}: {_to_str(count)} | {modules[module]}")

    remaining_parameters = sum([counts[k] for k in sorted_keys[topk:]])
    print(
        " " * 8,
        f"... and {len(counts)-topk} others accounting for {_to_str(remaining_parameters)} parameters",
    )
    return n_parameters



from typing_extensions import Annotated

import typer

from eztils import json


def dataclass_option(dataclass):
    def parser(x):
        if isinstance(x, dataclass):
            return x
        if isinstance(x, str):
            try:
                x = json.loads(x)
            except json.decoder.JSONDecodeError:
                raise typer.BadParameter(f"Could not parse {x} as a JSON string.")
            return dataclass(**x)
        if isinstance(x, dict):
            return dataclass(**x)

    return Annotated[dataclass, typer.Option(parser=parser)]  # create new type



class AttriDict(dict):  # type: ignore
    """
    A dictionary subclass that allows accessing keys as attributes.

    This class inherits from the built-in `dict` class and overrides the `__getattr__` and `__setattr__` methods to
    allow accessing keys as attributes. It also provides a `copy` method that returns a shallow copy of the instance.

    Note that attribute names that conflict with built-in dictionary methods or attributes cannot be used.

    Example usage:
    ```
    my_dict = AttriDict({'foo': 'bar'})
    print(my_dict.foo)  # prints 'bar'
    my_dict.baz = 'qux'
    print(my_dict['baz'])  # prints 'qux'
    ```

    :param dict: Initial dictionary to populate the instance with.
    :type dict: dict
    :raises AttributeError: If an attribute name conflicts with a built-in dictionary method or attribute.
    :return: An instance of the `AttriDict` class.
    :rtype: AttriDict
    """

    DICT_RESERVED_KEYS = list(vars(dict).keys())

    def __init__(self, *args, **kwargs):
        """
        Initializes an instance of the `AttriDict` class.

        :param args: Dictionaries to merge into the instance. e.g. ({}, {}, ..)
        :type args: dict
        :param kwargs: Key-value pairs to add to the instance.
        :type kwargs: Any
        """
        super().__init__(*args, **kwargs)
        self.__dict__ = self

    def __getattr__(self, attr):
        """
        Gets the value of a key as an attribute.

        :param attr: The name of the attribute to get.
        :type attr: str
        :return: The value of the key.
        :rtype: Any
        """
        if attr not in AttriDict.DICT_RESERVED_KEYS:
            return self.get(attr)
        return getattr(self, attr)

    def __setattr__(self, key, value):
        """
        Sets the value of a key as an attribute.

        :param key: The name of the key to set.
        :type key: str
        :param value: The value to set the key to.
        :type value: Any
        :raises AttributeError: If the key name conflicts with a built-in dictionary method or attribute.
        """
        if key == "__dict__":
            super().__setattr__(key, value)
            return
        if key in AttriDict.DICT_RESERVED_KEYS:
            raise AttributeError("You cannot set a reserved name as attribute")
        self.__setitem__(key, value)

    def __copy__(self):
        """
        Returns a shallow copy of the instance.

        :return: A shallow copy of the instance.
        :rtype: AttriDict
        """
        return self.__class__(self)

    def copy(self):
        """
        Returns a shallow copy of the instance.

        :return: A shallow copy of the instance.
        :rtype: AttriDict
        """
        return self.__copy__()
